(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.
Require Import Derived_multiple.
Require Import Derived_env.


(*************************************************************************)
(* N-ary universals *)
(*************************************************************************)

(* !a1:k1,...,aN:kN.t   :=   !a1:k1....!aN:kN.t *)
Fixpoint allsT ks t {struct ks} : typ :=
  match ks with
  | nil => t
  | k::ks' => allT k (allsT ks' t)
  end.

(* \a1:k1,...,aN:kN.e   :=   \a1:k1....\aH:kN.e *)
Fixpoint gensE ks e {struct ks} : exp :=
  match ks with
  | nil => e
  | k::ks' => genE k (gensE ks' e)
  end.

(* e (t1,...,tN)   :=   e t1 ... tN *)
Fixpoint instsE e ts {struct ts} : exp :=
  match ts with
  | nil => e
  | t::ts' => instsE (instE e t) ts'
  end.

(*
Lemma varsAT_alls : forall ks t,
  varsAT (allsT ks t) = varsAT t.
Proof.
  intros ks t.
  unfold allsT.
  induction ks.
    Case "ks = nil".
      reflexivity.
    Case "ks = k::ks'".
      unfold varsAT. fold varsAT.
      apply IHks.
Qed.
*)


(*************************************************************************)
(* Substitution *)
(*************************************************************************)

Lemma substAT_alls : forall a t' ks t,
  substAT a t' (allsT ks t) = allsT ks (substAT (shifts (length ks) a) t' t).
Proof.
  intros a t' ks t.
  generalize dependent a. generalize dependent t.
  induction ks; intros t a'; simpl allsT; simpl substAT.
    Case "ks = nil".
      reflexivity.
    Case "ks = k::ks'".
      rewrite IHks. rewrite shift_shifts. rewrite shifts_shift. reflexivity.
Qed.

Lemma openAT_alls : forall t' ks t,
  openAT t' (allsT ks t) = allsT ks (substAT (length ks) t' t).
Proof.
  intros t' ks t. unfold openAT.
  rewrite substAT_alls. rewrite shifts_bnd. rewrite plus_zero. reflexivity.
Qed.


Lemma substAE_gens : forall a t ks e,
  substAE a t (gensE ks e) = gensE ks (substAE (shifts (length ks) a) t e).
Proof.
  intros a t ks e.
  generalize dependent a. generalize dependent e.
  induction ks; intros e a'; simpl gensE; simpl substAE.
    Case "ks = nil".
      reflexivity.
    Case "ks = k::ks'".
      rewrite IHks. rewrite shift_shifts. rewrite shifts_shift. reflexivity.
Qed.

Corollary openAE_gens : forall t ks e,
  openAE t (gensE ks e) = gensE ks (substAE (length ks) t e).
Proof.
  intros t ks e. unfold openAE.
  rewrite substAE_gens. rewrite shifts_bnd. rewrite plus_zero. reflexivity.
Qed.

Lemma substXE_gens : forall x e' ks e,
  substXE x e' (gensE ks e) = gensE ks (substXE x e' e).
Proof.
  intros x e' ks e.
  generalize dependent x. generalize dependent e.
  induction ks; intros e x; simpl gensE; simpl substXE.
    Case "ks = nil".
      reflexivity.
    Case "ks = k::ks'".
      rewrite IHks. reflexivity.
Qed.

Corollary openXE_gens : forall e' ks e,
  openXE e' (gensE ks e) = gensE ks (openXE e' e).
Proof.
  intros e' ks e. unfold openXE. rewrite substXE_gens. reflexivity.
Qed.


Lemma substAE_insts : forall a t e ts,
  substAE a t (instsE e ts) = instsE (substAE a t e) (map (substAT a t) ts).
Proof.
  intros a t e ts.
  generalize dependent e.
  induction ts; intros e; simpl instsE; simpl substAE; simpl substAT.
    Case "ts = nil".
      reflexivity.
    Case "ts = t::ts'".
      rewrite IHts. reflexivity.
Qed.

Lemma openAE_insts : forall t e ts,
  openAE t (instsE e ts) = instsE (openAE t e) (map (openAT t) ts).
Proof.
  intros t e ts. unfold openAE, openAT. rewrite substAE_insts. reflexivity.
Qed.

Lemma opensAE_insts:
  forall (us : list typ) (e : exp) (ts : list typ),
  opensAE us (instsE e ts) = instsE (opensAE us e) (map (opensAT us) ts).
induction us as [ | u us].
intros e ts.
simpl.
rewrite (map_id typ ts).
reflexivity.
intros e ts.
simpl.
rewrite IHus.
rewrite substAE_insts.
rewrite map_map.
reflexivity.
Qed.

Lemma substXE_insts : forall x e' e ts,
  substXE x e' (instsE e ts) = instsE (substXE x e' e) ts.
Proof.
  intros x e' e ts.
  generalize dependent e.
  induction ts; intros e; simpl instsE; simpl substXE.
    Case "ts = nil".
      reflexivity.
    Case "ts = t::ts'".
      rewrite IHts. reflexivity.
Qed.

Lemma openXE_insts : forall e' e ts,
  openXE e' (instsE e ts) = instsE (openXE e' e) ts.
Proof.
  intros e' e ts. unfold openXE. apply substXE_insts.
Qed.


(*************************************************************************)
(* Local closure *)
(*************************************************************************)

Lemma lclosedT_alls : forall L ks t,
  (forall as', fresh L (length ks) as' -> lclosedT (opensAT (freeTs as') t)) ->
  lclosedT (allsT ks t).
Proof.
  intros L ks. generalize dependent L.
  induction ks; intros L t Ht; simpl allsT.
    replace t with (opensAT (freeTs nil) t).
      apply Ht. simpl fresh. trivial.
      unfold freeTs. simpl map. simpl opensAT. reflexivity.
    apply lclosedT_all with (L := L). intros a0 Fr.
      rewrite openAT_alls. apply IHks with (L := L \u {{a0}}). intros as' Frs.
        replace (length ks) with (length (freeTs as')).
          rewrite <- opensAT_cons.
            rewrite <- freeTs_cons. apply Ht. simpl fresh. split. apply Fr. apply Frs.
            apply lclosedT_var.
            apply lclosedTs_vars. rewrite freeTs_length. symmetry. eapply fresh_length. apply Frs.
Qed.

Lemma lclosedE_gens : forall L ks e,
  (forall as', fresh L (length ks) as' -> lclosedE (opensAE (freeTs as') e)) ->
  lclosedE (gensE ks e).
Proof.
  intros L ks. generalize dependent L.
  induction ks; intros L e He; simpl gensE.
    replace e with (opensAE (freeTs nil) e).
      apply He. simpl fresh. trivial.
      unfold freeEs. simpl map. simpl opensAE. reflexivity.
    apply lclosedE_gen with (L := L). intros a0 Fr.
      rewrite openAE_gens. apply IHks with (L := L \u {{a0}}). intros as' Frs.
        replace (length ks) with (length (freeTs as')).
          rewrite <- opensAE_cons.
            rewrite <- freeTs_cons. apply He. simpl fresh. split. apply Fr. apply Frs.
            apply lclosedT_var.
            apply lclosedTs_vars. rewrite freeTs_length. symmetry. eapply fresh_length. apply Frs.
Qed.

Lemma lclosedE_insts : forall e ts,
  lclosedE e ->
  lclosedTs ts ->
  lclosedE (instsE e ts).
Proof.
  intros e ts. generalize dependent e.
  induction ts; intros e He Hts; simpl instsE.
    apply He.
    apply IHts.
      apply lclosedE_inst.
        apply He.
        eapply lclosedTs_head. apply Hts.
      eapply lclosedTs_tail. apply Hts.
Qed.


(*************************************************************************)
(* Typing *)
(*************************************************************************)

Lemma ofT_alls : forall L D ks t,
  okD D ->
  (forall as', fresh L (length ks) as' ->
   ofT (D & as'~~ks) (opensAT (freeTs as') t) omK) ->
  ofT D (allsT ks t) omK.
Proof.
  intros L D ks t.
  generalize dependent t.
  generalize dependent D. generalize dependent L.
  induction ks; intros L D t HD Ht; simpl allsT in *.
    Case "ks = nil".
      assert (Ht' : ofT (D & nil~~nil) (opensAT (freeTs nil) t) omK).
        apply Ht. simpl fresh. trivial.
      apply Ht'.
    Case "ks = k :: ks'".
      apply ofT_all with (L := L \u dom D).
        intros a0 Fr. rewrite openAT_alls. apply IHks with (L := L \u {{a0}}).
          apply okD_var. apply HD. notin_solve.
          intros as' Frs.
            rewrite concat_assoc. rewrite <- concat_multi_cons.
            replace (length ks) with (length as').
              rewrite <- opensAT_cons_vars. apply Ht.
                simpl fresh. split. notin_solve. apply Frs.
              symmetry. eapply fresh_length. apply Frs.
Qed.

(*@crusso: move to fomega_structural as in branch *)
Lemma renamingAT_head : forall a1 a2 k1 D t k,
  a1 \notin dom D \u varsAT t ->
  a2 \notin dom D \u varsAT t ->
  ofT (D & a1~k1) (openAT a1 t) k ->
  ofT (D & a2~k1) (openAT a2 t) k.
Proof.
  intros a1 a2 k1 D t k Ha1 Ha2 Ht.
  destruct (a1 == a2). subst. apply Ht.
  assert (HD1 : okD (D & a1~k1)).
    eapply validityDT. apply Ht.
  assert (HD : okD D).
    inversion HD1. apply H1.
  assert (HD2 : okD (D & a2~k1)).
    apply okD_var. apply HD. notin_solve.
  assert (HD12 : okD (D & a1~k1 & a2~k1)).
    apply okD_var. apply HD1. rewrite dom_concat. rewrite dom_single. notin_solve.
  rewrite (openAT_as_substAT a1 a2).
    eapply substitutabilityAT.
      eapply weakeningDT_head. apply Ht. apply HD12.
      apply ofT_var. apply HD2. apply binds_head. notin_solve.
Qed.


Lemma ofT_alls_inversion : forall D ks t,
  ofT D (allsT ks t) omK ->
  exists L,
  (forall as', fresh L (length ks) as' ->
   ofT (D & as'~~ks) (opensAT (freeTs as') t) omK).
Proof.
  intros D ks; generalize D.
  clear D.
  induction ks; intros D t.
  Case "ks = nil".
     intros. 
     exists {}.
     intros.
     simpl length in H0.
     assert (RW := fresh_length _  _ _ H0).
     symmetry in RW.
     assert (RWas' := length_zero_is_nil _ _ RW).
     rewrite RWas'.
     apply H.
  Case "ks = a::ks".
     intro H.
     simpl allsT in  H.
     inversion H.
     pick_fresh_gen (L \u varsAT t ) a0.
     assert (FrL : a0 \notin L).
      notin_solve.
     assert (H2a := H2 a0 FrL).
     rewrite openAT_alls in H2a.
     assert (IHH2a := IHks _ _ H2a).
     destruct IHH2a as [L1 H4].
     exists (L1 \u {{a0}} \u dom(D)).
     intros.
     assert (RW1 := fresh_length _ _ _ H5).
     simpl length in RW1.
     symmetry in RW1.
     assert (H6 := length_succ_is_cons _ _ _ RW1).
     destruct H6 as [b [bs RWas']].
     rewrite RWas'. 
     rewrite concat_multi_cons.     
     unfold freeTs. simpl.
     fold (freeTs bs).
     rewrite substAT_opensAT.
     2: omega. 
     2: apply lclosedT_var. 
     2: apply lclosedTs_vars.
     assert (RWlen_free_bs : length (freeTs bs) = length ks).
      rewrite RWas' in RW1.
      injection RW1.
      intros RWlenbs.
      unfold freeTs.
      rewrite map_length.
      rewrite map_length.
      exact RWlenbs.
     assert (H6 : ofT (D & a0 ~ a & bs ~~ ks) (opensAT (freeTs bs) (substAT (length ks) a0 t)) omK).
       apply H4.
       simpl in H5.
       rewrite RWas' in H5.
       simpl fresh in H5.
       destruct H5 as [_ H6].
       fresh_solve.
     assert (H7 : ofT (D & b ~ a & bs ~~ ks) (substAT a0 b (opensAT (freeTs bs) (substAT (length ks) a0 t))) omK).
      eapply (substitutabilityAT ).
      rewrite concat_assoc.
      eapply weakeningDT.
      rewrite <- concat_assoc.
      apply H6.
      (*   okD (D & b ~ a & (a0 ~ a & bs ~~ ks)), plausible *)
      rewrite <- concat_assoc.
      apply okD_vars.
       apply okD_var.
         apply okD_var.
         apply (validityDT _ _ _ H).
          rewrite RWas' in H5.
          fresh_solve.
          simpl.
          apply notin_union_l.
          apply notin_singleton_swap.
          rewrite RWas' in H5.
          fresh_solve.
          assert ( H7 := validityDT _ _ _ H2a).
          inversion H7; auto.
          simpl.
          rewrite RWas' in H5.
          simpl fresh in H5.
          destruct H5 as [_ H7].
          fresh_solve.
          (* ofT (D & b ~ a & bs ~~ ks) b a, clearly *)
         rewrite concat_assoc.
         rewrite <- concat_multi_cons.
         apply ofT_vars.
          rewrite RWas' in H5.
          apply okD_vars.
           apply (validityDT _ _ _ H).
           rewrite <- union_assoc in H5.
           assert (H7 := fresh_union_r _ _ _ _ H5).
           destruct H7 as [_ H8].
           assert (H9 := fresh_union_r _ _ _ _ H8).
           firstorder.
        simpl.
        apply in_or_app.
        right.
        auto.
     rewrite <- concat_assoc.
     rewrite <- substAT_opensAT in H7.
     rewrite <- substAT_bnd_as_substAT_free in H7.
     rewrite <- substAT_opensAT.    
     rewrite RWlen_free_bs.
     apply H7.
     omega.
     apply lclosedT_var.
     apply lclosedTs_vars.
      eapply notin_varsAT_opensAT.
       rewrite RWas' in H5.
       simpl fresh in H5.
       destruct H5 as [_ H8].
       assert (H9 := fresh_resize _ _ _ H8).
       assert (H10 := fresh_union_r _ _ _ _ H9).
       destruct H10 as [H11 _].
       assert (H12 := fresh_union_r _ _ _ _ H11).
       destruct H12 as [H13 _].
         eapply H13.
      fresh_solve. (*  a0 \notin varsAT t, by our choice of a0! *) 
     rewrite RWlen_free_bs.
     omega.
     apply lclosedT_var.
     apply lclosedTs_vars.
Qed.
    

Lemma eqT_alls : forall L ks t t',
  (forall as', fresh L (length ks) as' -> eqT (opensAT (freeTs as') t) (opensAT (freeTs as') t')) ->
  eqT (allsT ks t) (allsT ks t').
Proof.
  intros L ks. generalize dependent L.
  induction ks; intros L t t' Ht; simpl allsT.
    Case "ks = nil".
      replace t with (opensAT (freeTs nil) t). replace t' with (opensAT (freeTs nil) t').
        apply Ht. unfold length, fresh. trivial.
        unfold freeTs, opensAT. simpl. reflexivity.
        unfold freeTs, opensAT. simpl. reflexivity.
    Case "ts = t::ts".
      eapply eqT_all. intros a0 Fr.
      rewrite openAT_alls. rewrite openAT_alls.
      apply IHks with (L := L \u {{a0}}). intros as' Frs.
      replace (length ks) with (length (freeTs as')).
      rewrite <- opensAT_cons. rewrite <- opensAT_cons.
        rewrite <- freeTs_cons. apply Ht. simpl fresh. split. apply Fr. apply Frs.
        apply lclosedT_var. apply lclosedTs_vars.
        apply lclosedT_var. apply lclosedTs_vars.
        rewrite freeTs_length. symmetry. eapply fresh_length. apply Frs.
Qed.


Lemma ofE_gens : forall L D G ks e t,
  okG D G ->
  (forall as', fresh L (length ks) as' ->
   ofE (D & as'~~ks) G (opensAE (freeTs as') e) (opensAT (freeTs as') t)) ->
  ofE D G (gensE ks e) (allsT ks t).
Proof.
  intros L D G ks e t HG He.
  unfold gensE, allsT.
  generalize dependent e. generalize dependent t. generalize dependent D. generalize dependent L.
  induction ks; intros L D HG t e He.
    Case "ks = nil".
      apply (He nil).
        unfold length, fresh. trivial.
    Case "ks = k::ks'".
      apply ofE_gen with (L := L \u dom D). apply HG.
        intros a0 Fr.
        rewrite openAE_gens. rewrite openAT_alls.
        apply IHks with (L := L \u {{a0}}).
          apply weakeningDG_head.
            apply HG.
            apply okD_var. eapply validityDG. apply HG. notin_solve.
          intros as' Frs.
          rewrite concat_assoc. rewrite <- concat_multi_cons.
          replace (length ks) with (length as').
            rewrite <- opensAE_cons_vars. rewrite <- opensAT_cons_vars.
              apply He. unfold fresh. simpl. split. notin_solve. apply Frs.
            symmetry. eapply fresh_length. apply Frs.
Qed.

Lemma ofE_insts : forall D G e ts ks t,
  ofE D G e (allsT ks t) ->
  length ks = length ts ->
  (forall t k, In (t, k) (combine ts ks) -> ofT D t k) ->
  ofE D G (instsE e ts) (opensAT ts t).
Proof.
  intros D G e ts ks t He Len Hts.
  generalize dependent e. generalize dependent t. generalize dependent ts.
  induction ks; intros ts Len Hts t e He; simpl allsT.
    Case "ks = nil".
      unfold length in Len. destruct ts; simpl instsE.
        SCase "ts = nil".
          unfold opensAT. apply He.
        SCase "ts = t'::ts'".
          inversion Len.
    Case "ks = k::ks'".
      simpl length in Len.
      destruct ts; simpl instsE.
        SCase "ts = nil".
          inversion Len.
        SCase "ts = t'::ts'".
          inversion Len.
          simpl opensAT.
          rewrite substAT_opensAT.
            apply IHks.
              apply H0.
              intros t1 k Elem.
                unfold combine in Hts. apply Hts. apply in_cons. apply Elem.
              replace (length ts) with (length ks). fold allsT. rewrite <- openAT_alls. eapply ofE_inst.
                simpl allsT in He. apply He.
                apply Hts. unfold combine. apply in_eq.
           omega.
           eapply regularityT. apply Hts. unfold combine. apply in_eq.
           intros t' Elem.
             assert (Len' := H0). symmetry in Len'.
             destruct (in_combine_l_inv _ _ ts ks t' Len' Elem).
             eapply regularityT. apply Hts. simpl combine. apply in_cons. apply H.
Qed.
