(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import Metatheory_Misc.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.
Require Import Derived_multiple.


(*************************************************************************)
(* Multiple mappings *)
(*************************************************************************)

Notation "xs ~~ ts" := (rev (combine xs ts))
  (at level 31, left associativity) : env_scope.


Lemma concat_multi_cons : forall (A : Set) (x : var) (t : A) xs ts,
  (x::xs)~~(t::ts) = x~t & xs~~ts.
Proof.
  intros A x t xs ts.
  unfold concat, combine. simpl. reflexivity.
Qed.

Lemma concat_multi_app : forall (A : Set) xs1 xs2 (ts1 ts2 : list A),
  length ts1 = length xs1 ->
  length ts2 = length xs2 ->
  (xs1 ++ xs2) ~~ (ts1++ts2) = xs1~~ts1 & xs2~~ts2.
Proof.
  intros A.
  induction xs1; intros xs2 ts1 ts2 Len1 Len2; destruct ts1; simpl length in Len1; inversion Len1.
    simpl. fold (empty (A := A)). rewrite <- empty_concat. reflexivity.
    simpl. rewrite IHxs1. 2: apply H0. 2: apply Len2.
      unfold concat. rewrite app_ass. reflexivity.
Qed.


Lemma dom_multi : forall (A : Set) xs (ts : list A),
  length ts = length xs ->
  dom (xs ~~ ts) = varsAs xs.
Proof.
  intro A.
  unfold varsAs.
  induction xs; intros ts Len; destruct ts; inversion Len; simpl combine; simpl rev.
    simpl. reflexivity.
    fold (concat ((a,a0)::nil) (xs~~ts)). rewrite dom_concat. simpl. rewrite union_empty_r.
      rewrite IHxs. reflexivity. apply H0.
Qed.


Lemma okD_vars : forall D as' ks,
  okD D ->
  fresh (dom D) (length ks) as' ->
  okD (D & as'~~ks).
Proof.
  intros D as' ks.
  generalize dependent as'. generalize dependent D.
  induction ks; intros D as' HD Fr.
    Case "ks = nil".
      replace as' with (nil (A := var)).
        unfold rev, combine. fold (empty (A := knd)). rewrite concat_empty. apply HD.
        unfold length in Fr. destruct as'. reflexivity. unfold fresh in Fr. destruct Fr.
    Case "ks = k::ks'".
      destruct as'.
        assert (Len : S (length ks) = length (nil (A := var))).
          apply fresh_length with (L := dom D). unfold length in Fr. apply Fr.
          inversion Len.
        unfold rev, combine.
          assert (Rew : forall A B : tenv, A & B = B ++ A). intros A B. unfold concat. reflexivity.
          rewrite <- Rew. rewrite <- concat_assoc.
          apply IHks.
            replace ((v,a)::nil) with (v~a). apply okD_var. apply HD. notin_solve. reflexivity.
            rewrite dom_concat. unfold fresh, length in Fr. destruct Fr.
              pattern dom at 2. unfold dom. rewrite union_empty_r. apply H0.
Qed.

Lemma okD_vars_inv : forall bs ks, forall D , length bs = length ks -> okD (D & bs ~~ ks) -> okD D /\ fresh (dom D) (length bs) bs.
Proof.
  induction bs as [ | b bs].
  simpl. auto.
  intros.
  simpl fresh.
  symmetry in H.
  assert (H1 := length_succ_is_cons _ _ _ H).
  destruct H1 as [k [ks' H2]].
  subst ks.
  rewrite -> concat_multi_cons in H0.
  rewrite <- concat_assoc in H0. 
  symmetry in H.
  inversion H as [lenbsislenks'].
  assert (IH := IHbs ks' _ lenbsislenks' H0).
  destruct IH.
  inversion H1.
  simpl fresh in H2.
  rewrite union_comm.
  firstorder.
Qed.


Lemma okG_vars : forall D G xs ts,
  okG D G ->
  (forall t, In t ts -> ofT D t omK) ->
  fresh (dom G) (length ts) xs ->
  okG D (G & xs~~ts).
Proof.
  intros D G xs ts.
  generalize dependent xs. generalize dependent G.
  induction ts; intros G xs HG Hts Fr.
    Case "ts = nil".
      replace xs with (nil (A := var)).
        unfold rev, combine. fold (empty (A := typ)). rewrite concat_empty. apply HG.
        unfold length in Fr. destruct xs. reflexivity. unfold fresh in Fr. destruct Fr.
    Case "ts = t::ts'".
      destruct xs.
        assert (Len : S (length ts) = length (nil (A := var))).
          apply fresh_length with (L := dom G). unfold length in Fr. apply Fr.
          inversion Len.
        unfold rev, combine.
          assert (Rew : forall A B : eenv, A & B = B ++ A). intros A B. unfold concat. reflexivity.
          rewrite <- Rew. rewrite <- concat_assoc.
          apply IHts.
            replace ((v,a)::nil) with (v~a). apply okG_var. apply HG. notin_solve. apply Hts. apply in_eq. reflexivity.
            intros t Elem. apply Hts. apply in_cons. apply Elem.
            rewrite dom_concat. unfold fresh, length in Fr. destruct Fr.
              pattern dom at 2. unfold dom. rewrite union_empty_r. apply H0.
Qed.


Lemma binds_tails : forall (A : Set) E xs ys x (y : A),
  ok (E & xs ~~ ys) ->
  In (x, y) (xs ~~ ys) ->
  binds x y (E & xs ~~ ys).
Proof.
  intros A E xs ys x y.
  induction (xs ~~ ys); intros HE Elem; subst.
    inversion Elem.
    destruct a; subst. env_fix. rewrite <- concat_assoc in *.
      destruct (v == x); subst.
      inversion Elem.
      inversion H. subst. apply binds_head.
        inversion HE. subst.
          assert (Elem' : x \in dom (E & l)).
            rewrite dom_concat. apply S.union_3.
            eapply in_dom. apply H.
          destruct (H4 Elem').
        unfold binds, Env.get. simpl.
        destruct (x == v).
          symmetry in e. destruct (n e).
          apply IHl.
            inversion HE. apply H1.
            inversion Elem. inversion H. subst. destruct n. reflexivity. apply H.
Qed.

Lemma ofT_vars : forall D as' ks a k,
  okD (D & as' ~~ ks) ->
  In (a, k) (as' ~~ ks) ->
  ofT (D & as' ~~ ks) a k.
Proof.
  intros D as' ks a k HD Elem.
  apply ofT_var.
    apply HD.
    apply binds_tails. eapply regularityD. apply HD. apply Elem.
Qed.

Lemma t_in_as : forall t (k : knd) as' ks,
  In (t, k) (combine (freeTs as') ks) ->
  exists a : var, t = a.
Proof.
  intros t k.
  induction as'; intros ks Elem; destruct ks; unfold freeTs in *; simpl map in *; simpl combine in *; inversion Elem.
    inversion H. subst. exists a. reflexivity.
    eapply IHas'. apply H.
Qed.

Lemma in_freeTs : forall (a : var) (k : knd) as' ks,
  In (varT a, k) (combine (freeTs as') ks) ->
  In (a, k) (combine as' ks).
Proof.
  intros a k.
  induction as'; intros ks Elem; destruct ks; unfold freeTs in *; simpl map in *; simpl combine in *; inversion Elem.
    inversion H. subst. apply in_eq.
    simpl In. right. apply IHas'. apply H.
Qed.

Lemma ofT_var_tail : forall D t k as' ks,
  length as' = length ks ->
  okD (D & as'~~ks) ->
  In (t, k) (combine (freeTs as') ks) ->
  ofT (D & as'~~ks) t k.
Proof.
  intros D t k.
  induction as'; intros ks Len HD Elem; destruct ks; simpl in Elem; inversion Elem.
    inversion H; subst. rewrite concat_multi_cons in *. rewrite <- concat_assoc in *. apply ofT_var.
      apply HD.
      apply binds_mid. eapply regularityD. apply HD.
    rewrite concat_multi_cons in *. rewrite <- concat_assoc in *. apply weakeningDT.
      simpl length in Len. inversion Len.
      apply IHas'.
        apply H1.
        destruct (okD_vars_inv _ _ _ H1 HD).
          inversion H0; subst.
          apply okD_vars. apply H5. simpl dom in H2. rewrite H1 in H2. fresh_solve.
        apply H.
      apply HD.
Qed.

