(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import Fomega_syntax.
Require Import Fomega_vars.


(*************************************************************************)
(* Values *)
(*************************************************************************)

Inductive value : exp -> Prop :=
  | value_trivE :
      value (trivE)
  | value_lamE : forall t e,
      lclosedE (lamE t e) ->
      value (lamE t e)
  | value_pairE : forall v1 v2,
      value v1 ->
      value v2 ->
      value (pairE v1 v2)
  | value_lftE : forall e t,
      value e ->
      lclosedT t ->
      value (lftE e t)
  | value_rgtE : forall e t,
      value e ->
      lclosedT t ->
      value (rgtE e t)
  | value_rollE : forall e t,
      value e ->
      lclosedT t ->
      value (rollE e t)
  | value_genE : forall k e,
      lclosedE (genE k e) ->
      value (genE k e)
  | value_packE : forall t1 e2 t,
      lclosedT t1 ->
      value e2 ->
      lclosedT t ->
      value (packE t1 e2 t)
.

Hint Constructors value.

(*************************************************************************)
(* Evaluation *)
(*************************************************************************)

Inductive step : exp -> exp -> Prop :=
  | step_app_1 : forall e1 e1' e2,
      step e1 e1' ->
      lclosedE e2 ->
      step (appE e1 e2) (appE e1' e2)
  | step_app_2 : forall v1 e2 e2',
      value v1 ->
      step e2 e2' ->
      step (appE v1 e2) (appE v1 e2')
  | step_app_beta : forall t e v,
      lclosedE (lamE t e) ->
      value v ->
      step (appE (lamE t e) v) (openXE v e)
  | step_pair_1 : forall e1 e1' e2,
      step e1 e1' ->
      lclosedE e2 ->
      step (pairE e1 e2) (pairE e1' e2)
  | step_pair_2 : forall v1 e2 e2',
      value v1 ->
      step e2 e2' ->
      step (pairE v1 e2) (pairE v1 e2')
  | step_fst : forall e e',
      step e e' ->
      step (fstE e) (fstE e')
  | step_fst_beta : forall v1 v2,
      value v1 ->
      value v2 ->
      step (fstE (pairE v1 v2)) v1
  | step_snd : forall e e',
      step e e' ->
      step (sndE e) (sndE e')
  | step_snd_beta : forall v1 v2,
      value v1 ->
      value v2 ->
      step (sndE (pairE v1 v2)) v2
  | step_lft : forall e e' t,
      step e e' ->
      lclosedT t ->
      step (lftE e t) (lftE e' t)
  | step_rgt : forall e e' t,
      step e e' ->
      lclosedT t ->
      step (rgtE e t) (rgtE e' t)
  | step_case : forall e1 e1' e2 e3,
      step e1 e1' ->
      lclosedE e2 ->
      lclosedE e3 ->
      step (caseE e1 e2 e3) (caseE e1' e2 e3)
  | step_case_beta_lft : forall v t e2 e3,
      value (lftE v t) ->
      lclosedE e2 ->
      lclosedE e3 ->
      step (caseE (lftE v t) e2 e3) (openXE v e2)
  | step_case_beta_rgt : forall v t e2 e3,
      value (lftE v t) ->
      lclosedE e2 ->
      lclosedE e3 ->
      step (caseE (rgtE v t) e2 e3) (openXE v e3)
  | step_roll : forall e e' t,
      step e e' ->
      lclosedT t ->
      step (rollE e t) (rollE e' t)
  | step_unroll : forall e e',
      step e e' ->
      step (unrollE e) (unrollE e')
  | step_unroll_beta : forall v t,
      value (rollE v t) ->
      step (unrollE (rollE v t)) v
  | step_inst : forall e e' t,
      step e e' ->
      lclosedT t ->
      step (instE e t) (instE e' t)
  | step_inst_beta : forall k e t,
      lclosedE (genE k e) ->
      lclosedT t ->
      step (instE (genE k e) t) (openAE t e)
  | step_pack : forall t1 e2 e2' t,
      lclosedT t1 ->
      step e2 e2' ->
      lclosedT t ->
      step (packE t1 e2 t) (packE t1 e2' t)
  | step_unpack : forall e1 e1' e2 (a : var) (x : var),
      step e1 e1' ->
      lclosedE (openAE a (openXE x e2)) ->
      step (unpackE e1 e2) (unpackE e1' e2)
  | step_unpack_beta : forall t1 v2 t e2,
      value (packE t1 v2 t) ->
      lclosedE (openAE t1 (openXE v2 e2)) ->
      step (unpackE (packE t1 v2 t) e2) (openAE t1 (openXE v2 e2))
.

Hint Constructors step.
