(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Export List.

Inductive occ : Set :=
  | free : var -> occ
  | bnd  : nat -> occ
.

Inductive knd : Set :=
  | omK   : knd							(* Omega *)
  | unitK : knd							(* 1 *)
  | arrK  : knd -> knd -> knd			(* k1->k2 *)
  | timK  : knd -> knd -> knd			(* k1*k2 *)
.

Inductive typ : Set :=
  | varT  : occ -> typ					(* a *)
  | unitT : typ							(* 1 *)
  | arrT  : typ -> typ -> typ			(* t1->t2 *)
  | timT  : typ -> typ -> typ			(* t1*t2 *)
  | plusT : typ -> typ -> typ			(* t1+t2 *)
  | recT  : knd -> typ -> typ			(* mu a:k.t *)
  | allT  : knd -> typ -> typ			(* !a:k.t *)
  | exT   : knd -> typ -> typ			(* ?a:k.t *)
  | trivT : typ							(* () *)
  | lamT  : knd -> typ -> typ			(* \a:k.t *)
  | appT  : typ -> typ -> typ			(* t1 t2 *)
  | pairT : typ -> typ -> typ			(* (t1,t2) *)
  | fstT  : typ -> typ					(* t.1 *)
  | sndT  : typ -> typ					(* t.2 *)
.

Inductive exp : Set :=
  | varE  : occ -> exp					(* x *)
  | trivE : exp							(* () *)
  | lamE  : typ -> exp -> exp			(* \x:t.e *)
  | appE  : exp -> exp -> exp			(* e1 e2 *)
  | pairE : exp -> exp -> exp			(* (e1,e2) *)
  | fstE  : exp -> exp					(* e.1 *)
  | sndE  : exp -> exp					(* e.2 *)
  | lftE  : exp -> typ -> exp			(* L e as t *)
  | rgtE  : exp -> typ -> exp			(* R e as t *)
  | caseE : exp -> exp -> exp -> exp	(* case e1 of L x => e2 | R x => e3 *)
  | rollE : exp -> typ -> exp			(* roll e as t *)
  | unrollE  : exp -> exp					(* unroll e *)
  | genE  : knd -> exp -> exp			(* \a:k.e *)
  | instE : exp -> typ -> exp			(* e t *)
  | packE : typ -> exp -> typ -> exp	(* (t1,e) as t2 *)
  | unpackE : exp -> exp -> exp			(* unpack (a,x) = e1 in e2 *)
.

Coercion free : var >-> occ.
Coercion bnd : nat >-> occ.
Coercion varT : occ >-> typ.
Coercion varE : occ >-> exp.

Notation tenv := (env knd).
Notation eenv := (env typ).
