(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import List.

(*************************************************************************)
(* Generic lemmas about sets *)
(*************************************************************************)

Lemma in_singleton_inv : forall x y,
  x \in {{y}} -> y \in {{x}}.
Proof.
  intros x y H.
  assert (Aux1 := in_singleton x y). destruct Aux1.
  assert (Aux2 := in_singleton y x). destruct Aux2.
  assert (Eq : x = y). apply H0. apply H.
  apply H3. symmetry. apply Eq.
Qed.

Lemma notin_singleton_inv : forall x y,
  x \notin {{y}} -> y \notin {{x}}.
Proof.
  intros x y H Elem.
  assert (Elem' : x \in {{y}}). apply in_singleton_inv. apply Elem.
  generalize Elem'.
  assert (Nelem : x \notin {{y}}). notin_solve.
  apply Nelem.
Qed.


(*************************************************************************)
(* Fresh variables *)
(*************************************************************************)

Lemma fresh_add : forall as1 as2 n1 n2 L,
  fresh L (n1 + n2) (as1 ++ as2) ->
  length as1 = n1 ->
  fresh L n1 as1 /\ fresh L n2 as2.
Proof.
  induction as1; intros as2 n1 n2 L H Len; subst; simpl length in *; simpl fresh in *; split.
    trivial.
    apply H.
    destruct H. split. apply H. eapply proj1. apply IHas1. apply H0. reflexivity.
    destruct H. eapply proj2. apply IHas1. eapply proj1. apply fresh_union_r. apply H0. reflexivity.
Qed.

(*************************************************************************)
(* Lists of variables *)
(*************************************************************************)

Definition varsAs as' := fold_right S.union S.empty (map S.singleton as').

Lemma varsAs_nil :
  varsAs (nil (A := var)) = {}.
Proof.
  unfold varsAs. simpl map. simpl fold_right. reflexivity.
Qed.

Lemma varsAs_cons : forall as' a,
  varsAs (a::as') = {{a}} \u varsAs as'.
Proof.
  unfold varsAs.
  induction as'; intro a'; simpl fold_right; reflexivity.
Qed.


Lemma fresh_single : forall as' a,
  a \notin varsAs as' ->
  fresh {} (length as') as' ->
  fresh {{a}} (length as') as'.
Proof.
  induction as'; intros a' Nelem Dis; simpl fresh.
    trivial.
    split; rewrite varsAs_cons in Nelem.
      apply notin_singleton_inv. notin_solve.
      simpl in Dis. destruct Dis.
        apply fresh_union_l.
          apply IHas'. notin_solve. fresh_solve.
          fresh_solve.
Qed.

Lemma fresh_concat_r : forall as1 as2 L,
  fresh L (length as1 + length as2) (as1 ++ as2) ->
  fresh L (length as1) as1 /\
  fresh (L \u varsAs as1) (length as2) as2.
Proof.
  induction as1; simpl; intros.
    split.
      trivial.
      rewrite varsAs_nil. rewrite union_empty_r. apply H.
    destruct H. split. split.
      notin_solve.
      eapply proj1. apply IHas1. apply H0.
      rewrite varsAs_cons. rewrite union_assoc. eapply proj2. apply IHas1. apply H0.
Qed.

Lemma fresh_concat_l : forall as1 as2 L,
  fresh L (length as1) as1 ->
  fresh (L \u varsAs as1) (length as2) as2 ->
  fresh L (length as1 + length as2) (as1 ++ as2).
Proof.
  induction as1; simpl; intros.
    rewrite varsAs_nil in H0. rewrite union_empty_r in H0. apply H0.
    destruct H. split.
      apply H.
      apply IHas1.
        apply H1.
        rewrite varsAs_cons in H0. rewrite union_assoc in H0. apply H0.
Qed.

Lemma fresh_partition : forall A (ks1 ks2 : list A) L as',
  fresh L (length ks1 + length ks2) as' ->
  exists as1, exists as2,
  as1 ++ as2 = as' /\
  length ks1 = length as1 /\
  length ks2 = length as2.
Proof.
  induction ks1; intros ks2 L as' Fr; simpl length in Fr.
    simpl fresh in Fr.
      exists (nil (A := var)). exists as'.
      split. reflexivity.
      split. reflexivity.
      eapply fresh_length. apply Fr.
    simpl in Fr. 
(*    rewrite plus_succ_l in Fr. *)
      destruct as'. simpl fresh in Fr. inversion Fr.
      destruct (IHks1 ks2 (L \u {{v}}) as') as [as1 [as2 [Eq [Len1 Len2]]]].
        simpl fresh in Fr. destruct Fr. apply H0.
      exists (v::as1). exists as2.
      split. rewrite <- Eq. symmetry. apply app_comm_cons.
      split. simpl length. f_equal. apply Len1.
      apply Len2.
Qed.


(*************************************************************************)
(* Environments *)
(*************************************************************************)

Lemma bind_concat_tail : forall (A : Set) E F x (y : A),
  binds x y F ->
  binds x y (E & F).
Proof.
  intros A E.
  induction F; intros x y H.
    inversion H.
    inversion H. destruct a; simpl in *. unfold binds, Env.get in *. destruct (x == v).
      apply H1.
      apply IHF. apply H1.
Qed.

Lemma in_dom : forall (A : Set) x (y : A) E,
  In (x, y) E ->
  x \in (dom E).
Proof.
  intros A x y.
  induction E; intro Elem; inversion Elem; subst; simpl dom.
    apply S.union_2. apply S.singleton_2. apply S.E.eq_refl.
    destruct a. subst. apply S.union_3. apply IHE. apply H.
Qed.
