(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import List.
Require Import Fomega.
Require Import Derived.
Require Import Modules_syntax.
Require Import Modules_semantic.


(*************************************************************************)
(* Environment well-formedness *)
(*************************************************************************)

Inductive wfG : tenv -> menv -> Prop :=
  | wfG_empty : forall C,
      okD C ->
      wfG C nil
  | wfG_var : forall C X s x G,
      wfG C G ->
      x \notin dom (transG G) ->
      ofT C (transS s) omK ->
      wfG C (G &' X~'s@x)
.

(*************************************************************************)
(* Matching *)
(*************************************************************************)

Inductive matchS : tenv -> sig -> sig -> list knd -> list typ -> exp -> Prop :=
  | matchS_enrich : forall C s s' ks ts f,
      length ks = length ts ->
      (forall t k, In (t, k) (List.combine ts ks) -> ofT C t k) ->
      enrichS C s (opensAS ts s') f ->
      matchS C s s' ks ts f

with enrichS : tenv -> sig -> sig -> exp -> Prop :=
  | enrichS_val : forall C t t',
      lclosedT t ->
      lclosedT t' ->
(*   rossberg: we should not need these:
      ofT C t omK ->  
      ofT C t' omK -> 
*)
      eqT t t' ->
      enrichS C (valS t) (valS t') (lamE (transS (valS t)) 0)
  | enrichS_typ : forall C t t' k,
      lclosedT t ->
      lclosedT t' ->
(*   rossberg: we should not need these:
      ofT C t k ->  
      ofT C t' k -> 
*)
      eqT t t' ->
      enrichS C (typS t k) (typS t' k) (lamE (transS (typS t k)) 0)
  | enrichS_sig : forall C q q' f f',
      enrichQ C q q' f ->
      enrichQ C q' q f' ->
      enrichS C (sigS q) (sigS q') (lamE (typT (transQ q) omK) (typE (transQ q') omK))
  | enrichS_str : forall C r r' re,
      distinct (labsR r) ->
      distinct (labsR r') ->
      enrichR C r r' re ->
      enrichS C (strS r) (strS r') (lamE (transS (strS r)) (strE re))
  | enrichS_fun : forall L C ks s q ks' s' q' ts f1 f2,
      (forall as', fresh L (length ks') as' ->
         matchS (C & as'~~ks')
           (opensAS (freeTs as') s')
           s ks (map (opensAT (freeTs as')) ts)
           (opensAE (freeTs as') f1)) ->
      (forall as', fresh L (length ks') as' ->
         enrichQ (C & as'~~ks')
           (opensAQ (map (opensAT (freeTs as')) ts) q)
           (opensAQ (freeTs as') q')
           (opensAE (freeTs as') f2)) ->
      enrichS C (funS ks s q) (funS ks' s' q')
        (lamE (transS (funS ks s q))
         (gensE ks'
          (lamE (transS s')
           (appE f2 (appE (instsE 1 ts) (appE f1 0))))))

with enrichR : tenv -> str -> str -> row exp -> Prop :=
  | enrichR_emp : forall C r,
(*       lclosedR r -> (*@crusso: could this just be lclosedT transR? *) *)
      lclosedT (transS (strS r)) ->
      enrichR C r empR nilR
  | enrichR_add : forall C l s1 r1 s2 r2 f re,
      field l s1 r1 ->
      enrichS C s1 s2 f ->
      enrichR C r1 r2 re ->
      enrichR C r1 (labR l s2 r2) (consR l (appE f (dotE (transR r1) l 0)) re)

with enrichQ : tenv -> qsig -> qsig -> exp -> Prop :=
  | enrichQ_ex : forall L C ks s ks' s' ts f e,
      (forall as', fresh L (length ks) as' ->
        opensAE (freeTs as') e = 
         packsE (map (opensAT (freeTs as')) ts) (opensAE (freeTs as') (appE f 0)) (opensAT (freeTs as') (exsT ks' (transS s')))) ->
      (forall as', fresh L (length ks) as' ->
          matchS (C & as'~~ks) (opensAS (freeTs as') s) s' ks' (map (opensAT (freeTs as')) ts) (opensAE (freeTs as') f)) ->
      enrichQ C (exQ ks s) (exQ ks' s')
        (lamE (transQ (exQ ks s))
         (unpacksE ks (transS s) 0 e))
(* The alternative definition of enrichQ below is more direct, but harder to reason about because ts is not lclosed. 
   In the above rule, we express the body of packsE below as an expression e
   satisfying:
       opensAE (freeTs as') e =
        opensAE (freeTs as') (packsE ts (appE f 0) (exsT ks' (transS s'))) = 
          packsE (map (opensAT (freeTs as')) ts) (opensAE (freeTs as') (appE f 0)) (opensAT (freeTs as') (exsT ks' (transS s')))
   which is the rewrite we need in our proof of soundness.
   We use a similar trick in the elaboration rule for projections.
 
with enrichQ : tenv -> qsig -> qsig -> exp -> Prop :=
  | enrichQ_ex : forall L C ks s ks' s' ts f,
      (forall as', fresh L (length ks) as' ->
         matchS (C & as'~~ks) (opensAS (freeTs as') s) s' ks' (map (opensAT (freeTs as')) ts) (opensAE (freeTs as') f)) ->
      enrichQ C (exQ ks s) (exQ ks' s')
        (lamE (transQ (exQ ks s))
         (unpacksE ks (transS s) 0
          (packsE ts (appE f 0) (exsT ks' (transS s')))))
*).

Scheme IImatchS := Induction for matchS Sort Prop
with IIenrichS := Induction for enrichS Sort Prop
with IIenrichR := Induction for enrichR Sort Prop
with IIenrichQ := Induction for enrichQ Sort Prop.
Combined Scheme IImatch from IImatchS, IIenrichS, IIenrichR, IIenrichQ.


(*************************************************************************)
(* Translation *)
(*************************************************************************)

Reserved Notation "C ; G |=T T ~: k ~> t" (at level 20).
Reserved Notation "C ; G |=E E ~: t ~> e" (at level 20).
Reserved Notation "C ; G |=S S ~> q" (at level 20).
Reserved Notation "C ; G |=D D ~> q" (at level 20).
Reserved Notation "C ; G |=M M ~: q ~> e" (at level 20).
Reserved Notation "C ; G |=B B ~: q ~> e" (at level 20).
Reserved Notation "C ; G |=P P ~: s ~> e" (at level 20).

Inductive elabT : tenv -> menv -> Typ -> knd -> typ -> Prop :=
  | elabT_typ : forall C G t k,
      wfG C G ->
      ofT C t k ->
      C; G |=T TypT t ~: k ~> t
  | elabT_path : forall C G P t k e,
      C; G |=P P ~: typS t k ~> e ->
      C; G |=T PathT P ~: k ~> t
  | elabT_pack : forall C G S q,
      C; G |=S S ~> q ->
      C; G |=T PackT S ~: omK ~> transQ q
where "C ; G |=T T ~: k ~> t" := (elabT C G T k t)

with elabE : tenv -> menv -> Exp -> typ -> exp -> Prop :=
  (* this is bogus (no inner paths), but does not really matter except for 1st-class modules *)
  | elabE_exp : forall C G e t,
      ofE C (transG G) e t ->
      C; G |=E ExpE e ~: t ~> e
  | elabE_path : forall C G P t e,
      C; G |=P P ~: valS t ~> e ->
      C; G |=E PathE P ~: t ~> e
  | elabE_pack : forall C G X S s' x ks s ts f,
      binds' X s' x G ->
      C; G |=S S ~> exQ ks s ->
      matchS C s' s ks ts f ->
      C; G |=E PackE X S ~: exsT ks (transS s) ~> packsE ts (appE f x) (exsT ks (transS s))
where "C ; G |=E E ~: t ~> e" := (elabE C G E t e)

with elabS : tenv -> menv -> Sig -> qsig -> Prop :=
  | elabS_path : forall C G P q e,
      C; G |=P P ~: sigS q ~> e ->
      C; G |=S PathS P ~> q
  | elabS_str : forall C G D q,
      C; G |=D D ~> q ->
      C; G |=S StrS D ~> q
  | elabS_fun : forall L C G X S1 S2 ks s q,
      C; G |=S S1 ~> exQ ks s ->
      (forall as' x, fresh L (length ks) as' -> x \notin L ->
       (C & as'~~ks);  (G &' X~'(opensAS (freeTs as') s)@x)  |=S S2
           ~> opensAQ (freeTs as') q) ->
      C; G |=S FunS X S1 S2 ~> exQ nil (funS ks s q) 
  | elabS_typeq : forall L C G S Xs ks1 k ks2 s s' T t,
      C; G |=S S ~> exQ (ks1++k::ks2) s ->
      C; G |=T T ~: k ~> t ->
      (forall as1 a as2, fresh L (length ks1 + length (k::ks2)) (as1 ++ a::as2) ->
         length as1 = length ks1 ->
         opensAS (freeTs (as1 ++ as2)) s' = substAS a t (opensAS (freeTs (as1 ++ a::as2)) s)) ->
      C; G |=S TypEqS S Xs T ~> exQ (ks1++ks2) s'
where "C ; G |=S S ~> q" := (elabS C G S q)

with elabD : tenv -> menv -> Dec -> qsig -> Prop :=
  | elabD_val : forall C G X T t,
      C; G |=T T ~: omK ~> t ->
      C; G |=D ValD X T ~> exQ nil (strS (labR X (valS t) empR))
  | elabD_typ : forall C G X K,
      wfG C G ->
      C; G |=D TypD X K ~> exQ (K::nil) (strS (labR X (typS 0 K) empR))
  | elabD_typeq : forall C G X T t k,
      C; G |=T T ~: k ~> t ->
      C; G |=D TypEqD X T ~> exQ nil (strS (labR X (typS t k) empR))
  | elabD_mod : forall C G X S ks s,
      C; G |=S  S ~> exQ ks s  ->
      C; G |=D ModD X S ~> exQ ks (strS (labR X s empR))
  | elabD_sigeq : forall C G X S q,
      C; G |=S S ~> q ->
      C; G |=D SigEqD X S ~> exQ nil (strS (labR X (sigS q) empR))
  | elabD_emp : forall C G ,
      wfG C G ->
      C; G |=D EmpD ~> exQ nil (strS empR)
  | elabD_seq : forall L C G D1 D2 ks1 r1 r1' ks2 r2,
      distinct (labsR r1) ->
      distinct (labsR r2) ->
      C; G |=D D1 ~> exQ ks1 (strS r1) ->
      (forall as' xs, fresh L (length ks1) as' -> fresh L (sizeR r1) xs ->
       (C & as'~~ks1); (G &' (opensAR (freeTs as') r1)@@xs) |=D D2
           ~> opensAQ (freeTs as') (exQ ks2 (strS r2))) ->
      (forall as1 as2, fresh L (length ks1 + length ks2) (as1 ++ as2) ->
         length as1 = length ks1 ->
         opensAR (freeTs (as1 ++ as2)) r1' = opensAR (freeTs as1) r1) ->
      diffR r1' r2 = r1' ->
      C; G |=D SeqD D1 D2 ~> exQ (ks1++ks2) (strS (seqR r1' r2))
  | elabD_incl : forall C G S q,
      C; G |=S S ~> q  ->
      C; G |=D InclD S ~> q
where "C ; G |=D D ~> q" := (elabD C G D q)

with elabM : tenv -> menv -> Mod -> qsig -> exp -> Prop :=
  | elabM_id : forall C G X s x,
      wfG C G ->
      binds' X s x G ->
      C; G |=M IdM X ~: exQ nil s ~> x
  | elabM_str : forall C G B q e,
      C; G |=B B ~: q ~> e ->
      C; G |=M StrM B ~: q ~> e
  | elabM_dot : forall C G M X ks r s e e',
      distinct (labsR r) ->
      C; G |=M M ~: exQ ks (strS r) ~> e ->
      field X s r ->
      (* e' = (packsE (bndTs (length ks)) (dotE (transR r) X 0) (exsT ks (transS s))) *)
      (forall as', length ks = length as' ->
         opensAE (freeTs as') e' = packsE (freeTs as') (opensAE (freeTs as') (dotE (transR r) X 0)) (exsT ks (transS s))) ->
      C; G |=M DotM M X ~: exQ ks s
           ~> unpacksE ks (transS (strS r)) e e'
  | elabM_fun : forall L C G X S M ks s q e,
      C; G |=S S ~> exQ ks s ->
      (forall as' x, fresh L (length ks) as' -> x \notin L ->
       (C & as'~~ks); (G &' X~'(opensAS (freeTs as') s)@x) |=M M ~: (opensAQ (freeTs as') q) ~> (openXE x (opensAE (freeTs as') e))) ->
      C; G |=M FunM X S M ~: exQ nil (funS ks s q) ~> gensE ks (lamE (transS s) e)
  | elabM_app : forall C G X1 X2 ks s q x1 s2 x2 ts f,
      wfG C G ->
      binds' X1 (funS ks s q) x1 G ->
      binds' X2 s2 x2 G ->
      matchS C s2 s ks ts f ->
      C; G |=M AppM X1 X2 ~: opensAQ ts q ~> appE (instsE x1 ts) (appE f x2)
  | elabM_seal : forall C G X S s' x ks s ts f,
      binds' X s' x G ->
      C; G |=S S ~> exQ ks s ->
      matchS C s' s ks ts f ->
      C; G |=M SealM X S ~: exQ ks s ~> packsE ts (appE f x) (exsT ks (transS s))
  | elabM_unpack : forall C G E S q e,
      C; G |=E E ~: transQ q ~> e ->
      C; G |=S S ~> q ->
      C; G |=M UnpackM E S ~: q ~> e
where "C ; G |=M M ~: q ~> e" := (elabM C G M q e)

with elabB : tenv -> menv -> Bind -> qsig -> exp -> Prop :=
  | elabB_val : forall C G X E t e,
      C; G |=E E ~: t ~> e ->
      C; G |=B ValB X E ~: exQ nil (strS (labR X (valS t) empR))
           ~> strE (consR X e nilR)
  | elabB_typ : forall C G X T t k,
      C; G |=T T ~: k ~> t ->
      C; G |=B TypB X T ~: exQ nil (strS (labR X (typS t k) empR))
           ~> strE (consR X (typE t k) nilR)
  | elabB_mod : forall C G X M ks s e e',
      C; G |=M M ~: exQ ks s ~> e ->
      (* e' = (packsE (bndTs (length ks))
               (strE (consR X (varE 0) nilR))
               (exsT ks (transS (strS (labR X s empR))))) *)
      (forall as', length ks = length as' ->
       opensAE (freeTs as') e' =
              (packsE (freeTs as')
               (strE (consR X (varE 0) nilR))
               (exsT ks (transS (strS (labR X s empR)))))) ->
      C; G |=B ModB X M ~: exQ ks (strS (labR X s empR))
           ~> unpacksE ks (transS s) e e'
  | elabB_sig : forall C G X S q,
      C; G |=S S ~> q ->
      C; G |=B SigB X S ~: exQ nil (strS (labR X (sigS q) empR))
           ~> strE (consR X (typE (transQ q) omK) nilR)
  | elabB_emp : forall C G,
      wfG C G ->
      C; G |=B EmpB ~: exQ nil (strS empR) ~> strE nilR
  | elabB_seq : forall L C G B1 B2 ks1 r1 r1' e1 ks2 r2 e2 es' e',
      distinct (labsR r1) ->
      distinct (labsR r2) ->
      C; G |=B B1 ~: exQ ks1 (strS r1) ~> e1 ->
      (forall as1 xs, fresh L (length ks1) as1 -> fresh L (sizeR r1) xs ->
       (C & as1~~ks1); (G &' (opensAR (freeTs as1) r1)@@xs) |=B B2
            ~: opensAQ (freeTs as1) (exQ ks2 (strS r2)) ~> opensXE (freeEs xs) (opensAE (freeTs as1) e2)) ->
      (forall as1 as2, fresh L (length ks1 + length ks2) (as1 ++ as2) ->
         length as1 = length ks1 ->
         opensAR (freeTs (as1 ++ as2)) r1' = opensAR (freeTs as1) (diffR r1 r2)) ->
      (* es' = let rt1 := transR r1 in
              let re1 := mapiR (fun l _ => dotE rt1 l 0) rt1 in
              fieldsR re1 *)
      (forall x1 as1, length ks1 = length as1 ->
       map (openXE x1) (map (opensAE (freeTs as1)) es') =
              let rt1 := transR (opensAR (freeTs as1) r1) in
              let re1 := mapiR (fun l _ => dotE rt1 l x1) rt1 in
              fieldsR re1) ->
      (* e' = let rt1' := transR r1' in
              let rt2 := transR r2 in
              let re1' := mapiR (fun l _ => dotE rt1 l (S (length ks2))) rt1' in
              let re2 := mapiR (fun l _ => dotE rt2 l 0) rt2 in
               (packsE (bndTs (length (ks1++ks2)))
               (strE (catR re1' re2))
               (exsT (ks1++ks2) (transS (strS (seqR r1' r2))))) *)
      (forall x1 x2 as1 as2, length ks1 = length as1 -> length ks2 = length as2 ->
       openXE x2 (substXE (S (length ks2)) x1 (opensAE (freeTs (as1++as2)) e')) =
              let rt1 := transR (opensAR (freeTs as1) r1) in
              let rt1' := transR (opensAR (freeTs (as1++as2)) r1') in
              let rt2 := transR (opensAR (freeTs (as1++as2)) r2) in
              let re1' := mapiR (fun l _ => dotE rt1 l x1) rt1' in
              let re2 := mapiR (fun l _ => dotE rt2 l x2) rt2 in
              (packsE (freeTs (as1++as2))
               (strE (catR re1' re2))
               (exsT (ks1++ks2) (transS (strS (seqR r1' r2)))))) ->
      C; G |=B  SeqB B1 B2 ~: exQ (ks1++ks2) (strS (seqR r1' r2))
           ~> let rt1 := transR r1 in
              unpacksE ks1 (transS (strS r1)) e1
              (unpacksE ks2 (transS (strS r2)) (letsE (fieldsR rt1) es' e2) e')
  | elabB_incl : forall C G M q e,
      C; G |=M M ~: q ~> e ->
      C; G |=B InclB M ~: q ~> e
where "C ; G |=B B ~: q ~> e" := (elabB C G B q e)

with elabP : tenv -> menv -> Path -> sig -> exp -> Prop :=
  | elabP_path : forall (* L *) C G M  s ks e,
      C; G |=M M ~: exQ ks s ~> e ->
      ofT C (transS s) omK ->
      C; G |=P ModP M ~: s ~> unpacksE ks (transS s) e 0
where "C ; G |=P P ~: s ~> e" := (elabP C G P s e)
.


Scheme IIelabT := Induction for elabT Sort Prop
with IIelabE := Induction for elabE Sort Prop
with IIelabS := Induction for elabS Sort Prop
with IIelabD := Induction for elabD Sort Prop
with IIelabM := Induction for elabM Sort Prop
with IIelabB := Induction for elabB Sort Prop
with IIelabP := Induction for elabP Sort Prop.
Combined Scheme IIelab from IIelabT, IIelabE, IIelabS, IIelabD, IIelabM, IIelabB, IIelabP.
